use std::{
    cmp::{Ordering, Reverse},
    collections::HashMap,
    sync::LazyLock,
    usize,
};

use super::{dealt_hand::DealtHand, hand_kind::HandKind};

#[derive(Debug, PartialEq, Eq)]
pub struct CategorizedHand {
    kind: HandKind,
    rest: DealtHand,
}

impl CategorizedHand {
    pub fn rest(&self) -> &DealtHand {
        &self.rest
    }
    pub fn score(&self) -> usize {
        self.rest().score()
    }
}

impl Ord for CategorizedHand {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        match self.kind.cmp(&other.kind) {
            std::cmp::Ordering::Equal => compare_along_str(self.rest().hand(), other.rest().hand()),
            not_equal_ordering => not_equal_ordering,
        }
    }
}

impl PartialOrd for CategorizedHand {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

static LABEL_ORDERING: LazyLock<HashMap<char, usize>> = LazyLock::new(|| {
    let numbers = (0..9)
        .into_iter()
        .map(|number| std::char::from_digit(number, 10).unwrap());
    let letters = ['T', 'J', 'Q', 'K', 'A'];
    numbers
        .chain(letters.into_iter())
        .into_iter()
        .enumerate()
        .map(|(index, character)| (character, index))
        .collect()
});
//A hand consists of five cards labeled one of A, K, Q, J, T, 9, 8, 7, 6, 5, 4, 3, or 2.
//The relative strength of each card follows this order, where A is the highest and 2 is the lowest.
fn compare_along_str(left: &str, right: &str) -> Ordering {
    for (left_char, right_char) in left.chars().zip(right.chars()) {
        let left_rank_value = LABEL_ORDERING.get(&left_char);
        let right_rank_value = LABEL_ORDERING.get(&right_char);
        let current_ordering = left_rank_value.cmp(&right_rank_value);
        if current_ordering != Ordering::Equal {
            return current_ordering;
        }
    }
    left.len().cmp(&right.len())
}

impl From<DealtHand> for CategorizedHand {
    fn from(value: DealtHand) -> Self {
        let mut count_of_same: HashMap<char, usize> = HashMap::with_capacity(10);
        for next_char in value.hand().chars() {
            count_of_same
                .entry(next_char)
                .and_modify(|count| *count = *count + 1)
                .or_insert(1);
        }
        let values = {
            let mut to_sort: Vec<usize> =
                count_of_same.into_iter().map(|(_, count)| count).collect();
            to_sort.sort_by_key(|&count| Reverse(count));
            to_sort
        };

        let max = values.first().unwrap();
        let kind = match max {
            5 => HandKind::Five,
            4 => HandKind::Four,
            _ => {
                let second = values.get(1).unwrap();
                match (max, second) {
                    (3, 2) => HandKind::FullHouse,
                    (3, _) => HandKind::Three,
                    (2, 2) => HandKind::TwoPair,
                    (2, 1) => HandKind::OnePair,
                    _ => HandKind::HighCard,
                }
            }
        };

        CategorizedHand { kind, rest: value }
    }
}

#[cfg(test)]
mod testing {
    use std::cmp::Ordering;

    use crate::solutions::day7::{
        categorized_hand::CategorizedHand, dealt_hand::DealtHand, hand_kind::HandKind,
    };

    #[test]
    fn from_dealt_hand_to_categorized_hand() {
        fn assert_case(given: DealtHand, expected: HandKind) {
            let expected_hand = CategorizedHand {
                kind: expected,
                rest: given.clone(),
            };
            let actual = given.into();
            assert_eq!(expected_hand, actual);
        }

        assert_case(DealtHand::new("AAAAA", 20), HandKind::Five);
        assert_case(DealtHand::new("AA8AA", 20), HandKind::Four);
        assert_case(DealtHand::new("23332", 20), HandKind::FullHouse);
        assert_case(DealtHand::new("TTT98", 20), HandKind::Three);
        assert_case(DealtHand::new("23456", 20), HandKind::HighCard);
        assert_case(DealtHand::new("23432", 20), HandKind::TwoPair);
        assert_case(DealtHand::new("A23A4", 20), HandKind::OnePair);
    }

    #[test]
    fn correct_cmp() {
        fn assert_case(left: DealtHand, right: DealtHand, expected: Ordering) {
            let (left, right): (CategorizedHand, CategorizedHand) = (left.into(), right.into());
            let actual = left.cmp(&right);
            assert_eq!(expected, actual, "Left: {:?}\nRight: {:?}", left, right);
        }

        assert_case(
            DealtHand::new("AAAAA", 20),
            DealtHand::new("BBBBA", 30),
            Ordering::Greater,
        );
        assert_case(
            DealtHand::new("12345", 20),
            DealtHand::new("BBBBA", 30),
            Ordering::Less,
        );
        assert_case(
            DealtHand::new("12345", 20),
            DealtHand::new("28495", 30),
            Ordering::Less,
        );
        assert_case(
            DealtHand::new("QQQJA", 20),
            DealtHand::new("T55J5", 30),
            Ordering::Greater,
        );
        //KK677 and KTJJT are both two pair.
        //Their first cards both have the same label,
        //but the second card of KK677 is stronger (K vs T), so KTJJT gets rank 2 and KK677 gets rank 3.
        assert_case(
            DealtHand::new("KK677", 20),
            DealtHand::new("KTJJT", 30),
            Ordering::Greater,
        );
    }

    #[test]
    fn sort_example_correctly() {
        //32T3K is the only one pair and the other hands are all a stronger type, so it gets rank 1.
        //
        //KK677 and KTJJT are both two pair.
        //Their first cards both have the same label,
        //but the second card of KK677 is stronger (K vs T), so KTJJT gets rank 2 and KK677 gets rank 3.
        //
        //T55J5 and QQQJA are both three of a kind.
        //QQQJA has a stronger first card, so it gets rank 5 and T55J5 gets rank 4.

        let mut input: Vec<_> = vec![
            DealtHand::new("32T3K", 765),
            DealtHand::new("T55J5", 684),
            DealtHand::new("KK677", 28),
            DealtHand::new("KTJJT", 220),
            DealtHand::new("QQQJA", 483),
        ]
        .into_iter()
        .map(CategorizedHand::from)
        .collect();

        input.sort();

        insta::assert_debug_snapshot!(input);
    }
}
